; Real-Mode code & data for the TUNES BIOS-based API
; see ~/info/BIOS-API.nfo


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
GoRealSEG	=	RMPARTCS



rmpart

__dataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Variables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


RM_Target:
	RM_Target_ofs:	.word	Orig_RM_Target_ofs
	RM_Target_seg:	.word	Orig_RM_Target_seg

PM_Target:
	PM_Target_ofs:	.long	Orig_PM_Target_ofs
	PM_Target_seg:	.word	Orig_PM_Target_seg

PM_GDT_address:		; Length & Base address of protected-mode GDT
	PM_GDT_lim:	.word	Orig_GDT_Len-1
	PM_GDT_base:	.long	Orig_GDT

RM_SYSTACK:		; Length & Base address of protected-mode GDT
RM_SYSTACK_ofs:		.word	0 ; ?	; Saved address for R-mode stack
RM_SYSTACK_seg:		.word	0 ; ?	; Saved address for R-mode stack


;;;;;;;;;;;;;;; Standard objects (their low_level implementation) ;;;;;;;;;;;;

;;; Stdin as a byte per byte input ;;;
__stdin_param__:	.word	0 ; ?
__stdin_Getc__:		.word	cGetc
__stdin_Flush__:	.word	False
__stdin_EOF__:		.word	False
;;; Stdout as a byte per byte output ;;;
__stdout_param__:	.word	0 ; ?
__stdout_Putc__:	.word	cPutc
__stdout_Flush__:	.word	False
__stdout_FULL__:	.word	False
})



__udataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Uninitialized Data ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

rm_ss_sp	= allot(ru,0)
rm_sp		= allot(ru,2)
rm_ss		= allot(ru,2)
RMPartStackBot	= allot(ru,256*2)
RMPartStackTop	= allot(ru,0)

})


__codeseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;; Standard Routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Puts:
; Adapted from MicroSoft's own boot sector
; DS:SI=ASCIIZ string to print
; AX spoilt, SI points just after the string-ending zero, 7 put into BX.
	cld
	lodsb
	or	al,al
	jz	Return
	push	#Puts
;	jmps	Putc			; (!) Putc must follow.

Putc:
	push	si
	mov	si,[__stdout_param__]
	call	[__stdout_Putc__]
	pop	si
	ret

Getc:
	push	si
	mov	si,[__stdout_param__]
	call	[__stdout_Putc__]
	pop	si
	ret

cGetc:
	clr(ah)
	int #0x16
	ret

Beep:
; Echo a BEEP to the the Console
	mov	al,#7
	;jmps	cPutc

cPutc:
; Outputs a character to console
; spoils AX and DI, keeps si
	pusha
	push	es
	mov	ah,#0xE	; function Teletype
	mov	bx,#7	; attribute
	int	#0x10	; Video BIOS call
	pop	es
	popa
Return:	ret

; the true condition is "c"; false is "nc"
True:
	stc
	ret

False:
	clc
	ret
})


zzz({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; DEBUGGING AID ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
__stringseg__({
  RegsMsg:	.ascii	"DI   SI   BP   SP   "
		.ascii	"BX   DX   CX   AX   "
		.ascii	"FLGS CS   DS   ES   "
		.ascii	"SS   IP   STK1 STK2"
		.byte	13,10,0
  ERegsMsg1:	.ascii	"EDI      ESI      EBP      ESP      "
		.ascii	"EBX      EDX      ECX      EAX      EFLAGS"
		.byte	13,10,0
  ERegsMsg2:	.ascii	"CS   DS   ES   SS   IP   STK1 STK2 STK3 "
		.ascii	"STK4 STK5 STK6 STK7 STK8 STK9 STKA STKB"
		.byte	13,10,0
})
__codeseg__({

DumpERegs:
	push ss
	push es
	push ds
	push cs
	pushfd
	pushad
	mov si,#ERegsMsg1
	PUTS
	push ss
	pop es
	push cs
	pop ds
	mov di,sp
	mov cx,#9
	call PutLZone
	mov si,#ERegsMsg2
	PUTS
	mov cx,16
	call PutWZone
	popad
	popfd
	pop es	;no pop cs above 8086
	pop ds
	pop es
	pop ss
 Return2:
	ret

DumpRegs:
	push ss
	push es
	push ds
	push cs
	pushf
	pusha
	mov si,#RegsMsg
	PUTS
	mov di,sp
	push ss
	pop es
	push cs
	pop ds
	mov cx,#16
	call PutWZone
	call PrCRLF
	popa
	popf
	pop es	;no pop cs above 8086
	pop ds
	pop es
	pop ss
	ret

PutLZone_:
	call	Space
PutLZone:
	seg es
	mov eax,[di]
	add di,#4
	call PutHexDword
	loop PutLZone_
	ret

PutWZone_:
	call	Space
PutWZone:
	seg es
	mov ax,[di]
	inc di
	inc di
	call	PutHexWord
	loop PutWZone_
	ret

PutZone_:
	call	Space
PutZone:
	seg es
	mov al,[di]
	inc di
	call	PutHexByte
	jcxz	Return2
	call	Space
	loop PutZone_
	ret

PutAddr:
;; Prints the contents of es:di
	push	di
	mov	ax,es
	call	PutHexWord
	mov	al,':
	call	Putc
	pop	di
	mov	ax,di
	jmp	PutHexWord

PutHexDword:
	rol	eax,#16
	call	PutHexWord
	rol	eax,#16
;	jmp	PutHexWord

PutHexWord:
; Prints the contents of AX in hexa.
; everything but F saved
		xchg	al,ah
		call	PutHexByte
		xchg	al,ah
;		jmp	PutHexByte

})
})


__codeseg__({

PutHexByte:
; Prints the contents of AL in hexa.
; saves AX.
	push	cx
	mov	cl,#4
	rol	al,cl
	call	PutHexNybble
	rol	al,cl
	pop	cx
;	jmp	PutHexNybble

PutHexNybble:
; Prints to BIOS console the hexa nybble in low AL
; saves AX, puts 7 into BX
	push	ax
	and	al,0xF
	add	al,#'0
	cmp	al,#10+'0
	jb	do_putnyb
	add	al,#'A-'0-10
 do_putnyb:
	PUTC
	pop	ax
	ret

forget({
sPutc:
	stosb
	ret

sGetc:
	lodsb
	ret
})

PrCRLF:
	PUTC(#13)
	mov al,#10
	br Putc

zzz({
Space:
	mov al,#' 
	br Putc
})
})


__codeseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; I/O subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Delay:			;;; Delay between I/O, as in Linux.
	jmp DelayRet
DelayRet:
	ret

RM_Disable_NMI:
	LibInline(Disable_NMI)
	ret

RM_Enable_NMI:
	LibInline(Enable_NMI)
	ret

Reset_Copro:
	clr(ax)
	out	0xF0,al
	call	Delay
	out	0xF1,al
	ret

LibProc(KB_Ready,Timeout_8042)

; What to do in case of time out from 8042
Timeout_8042:
		.byte	0x68		; push immediate value of timeout exc.
 Time_Out_8042:	.word	KB_Ready	; default behaviour: retry indefinitely
		ret

Disable_A20:
	mov ah,#set_A20_OFF
	jmp Set_A20

Enable_A20:
	mov ah,#set_A20_ON
	;jmp Set_A20

LibProc(Set_A20)
})


__codeseg__({
;;;;;;;;;;;;;;;;;;;;;;; Real to Protected Mode switcher ;;;;;;;;;;;;;;;;;;;;;;;
DoCallINT:
; Call the Interrupt
		.byte	0xCD		; INT instruction
INT_NUM:	.byte	Orig_INT_NUM

Save_State:
; - Saves all register for use by Restore_State, except return address,
; which is left to Halt
; - jumps to protected mode
	pushf
	push cs
	push #Halt
	pushad
	clr(ax)
	push ax
	push es
	push ax
	push ds
	mov edi,esp
	mov si,ss

GoProtect:
; PREcondition
; * cs points to a segment so that RM_Data is at same offset as from API file
; POSTcondition
; * A20 line is enabled
; * all hardware interrupts are disabled
; * Processor enters Protected mode
; * Processor jumps into 32-bit far address contained in
; * eax,ebx,ecx,edx may be scrambled
; * ds,es,ss,fs,gs unchanged (internal values also)
; * esp,ebp,esi,edi unchanged

	zwoz({{"Going PM, Target="}},[PM_Target_seg],{":"})
	zdwoz({},[PM_Target_ofs],{CRLF})
	cli
	;__("Disabling NMI",CRLF)
	call	RM_Disable_NMI
	;__("Enabling A20",CRLF)
	call	Enable_A20
	zdwoz({"Loading GDT at "},[PM_GDT_base],{CRLF})
	lgdt	[PM_GDT_address]
	mov	eax,cr0		; 32 bit msw
	and	eax,#0x1FFAFFFF	; clear PG, CD, NW, AM, WP flags
	or	al,#1		; protected mode (PE) bit
	SST({"Hit any key to go PM"},CRLF)
	test	byte [GP_CR0],#1
	mov	cr0,eax		; This is it!
 GP_CR0:
	seg cs
	jmpi dword [PM_Target]



GoReal:
; PREconditions:
; * Interrupts are disabled
; * (!) cs points to a segment such that the PM Data area is at the
;  same offset as it is from the API file (that's klugdy) (FIXME!)
; POSTconditions:
; (just) do the switching to Real mode and jump to 16-bit far [RM_Target]
; NOTE:
; I had problems having this code in a 32 bit segment,
; as the final jump seemed to get nowhere (or get it in real 32-bit mode ?)
	;IF	__DEBUG__
	;mov eax,ALLMEMDS
	;mov fs,ax
	;ENDIF
	;STOP(fs,0xB8000,1,13,'G,0xC)
	mov	eax,cr0
	and	eax,#0x1ffafffe	; clear PG, CD, NW, AM, WP, and *PE* flags
	test	byte [GR_CR0],#1 ; put follwing instructions in cache...
	mov	cr0,eax		; Back to the Real World ):)
   GR_CR0:
	;STOP(fs,0xB8000,2,13,'o,0xC)
	flushinstr
	;STOP(fs,0xB8000,3,13,'R,0xC)
	seg cs
	jmpi word [RM_Target]

})



__codeseg__({
;;;;;;;;;;;;;;;;;;; Calling BIOS services from P-mode ;;;;;;;;;;;;;;;;;;;;;;;;
Restore_State:
; Parameters:
; - es:esp real stack pointer in si:edi
; - on this stack, registers pushed, to be recovered by
;  pop ds,pop 0,pop es, pop 0, popad, iret
; - fs and gs are unchanged...
	mov	es,si
	mov	esp,edi
	pop	ds
	pop	ax ; zero padding
	pop	es
	pop	ax
	popad

Ret_From_PM:
	iret			; !

Halt:
	hlt
	jmp Halt
})



__codeseg__({
;;;;;;;;;;;;;;;;;;; Hook from hardware IRQ to BIOS handlers ;;;;;;;;;;;;;;;;;;
;;; (when MOOSE I/O control is disabled)
RM_IRQ:
	int #0x08
	iret
	int #0x09
	iret
	int #0x0A
	iret
	int #0x0B
	iret
	int #0x0C
	iret
	int #0x0D
	iret
	int #0x0E
	iret
	int #0x0F
	iret
	int #0x70
	iret
	int #0x71
	iret
	int #0x72
	iret
	int #0x73
	iret
	int #0x74
	iret
	int #0x75
	iret
	int #0x76
	iret
	int #0x77
	iret
})
