;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; kernel.mS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;; i386 specific definitions for the LLL kernel ;;;;;;;;;;;;;;;;;;

include(as86.mS)
include(version)
include(magic.m4)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Structures ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Struc({AS},{
; Signature_Struc
  _long_(Magic_Number)		; magic number from magic file
  _byte_(Object_Name,20)
  _byte_(Version,8)
})

Struc({AH},{
; API_Header_Struc
; size is 32. offset is 32 from beginning of API file
; (32 first bytes are signature and version)
  _long_(TotalLength)		; total length of the file
  _long_(EntryPoint)		; address of object to return
  _long_(Offset)		; offset where to load it
  _long_(HeaderSize)		; size of header fields in file
  _long_(ExpandedSize)		; size when expanded to memory
  _long_()			; fill in
  _long_(CheckSum)		; checksum
  _long_(LongCheckSum)		; long checksum
})


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Some values ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
TRUE		=	-1	; Logical True value
FALSE		=	 0	; Logical False value

BYTEBITS	=	 8	; bits in a byte (smallest addressable value)

CHARABITS	=	 0	; bits of char in address
CELLABITS	=	 2	; bits of cell in address
PAGEABITS	=	12	; bits of page in address


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Deduced values ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CHARSIZE	=	(1 << CHARABITS)	; size of a char in bytes
CELLSIZE	=	(1 << CELLABITS)	; size of a cell in bytes
PAGESIZE	=	(1 << PAGEABITS)	; size of a page in bytes
CHARMASK	=	~ (CHARSIZE - 1)	; mask of chars in address
CELLMASK	=	~ (CELLSIZE - 1)	; mask of cells in address
PAGEMASK	=	~ (PAGESIZE - 1)	; mask of pages in address
BITSPERCHAR	=	CHARSIZE*BYTEBITS	; bits in a char
BITSPERCELL	=	CELLSIZE*BYTEBITS	; bits in a cell


;;;;;;;;;;;;;;;;;;;;;;;;; IMPLEMENTATIONS OF "NEXT" ;;;;;;;;;;;;;;;;;;;;;;;;;;

define({PC},		esi)	; program counter
define({DSP},		edi)	; data stack pointer
DSINC		=	+4	; increment for data stack
define({RSP},		ebp)	; RSP is the return stack pointer (growing up)
RSINC		=	-4	; increment for return stack
define({TOS},		ebx)	; buffer for the top of stack
define({R0},		eax)	; a scratch register
define({R0L},		al)	; its lowest byte
define({R0H},		ah)	; its second byte
define({INSTRTBL},	InstrT)	; table for bytecode threading


define({__ENTER__},{
	add PC,4
})
define({__NEXT__},{
	jmp [PC]
})


;;;;;;;;;;;;;;;;;;;;;;;;;; Defining FORTH words ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; defining a counted string
define({DefCString},
{__stringseg__({
locals({strlen},{strbeg},{strend},{
strbeg:
	.byte 	strlen
	defb($@)
strend:
strlen	=	strend-strbeg-1
})})})

;; defining a FORTH word


define({DefWord},{
ifelse({$2},,{$1},{$2}):
})

; flags for FORTH words:
;FW_IMMED	= 0x80	; word is immediate
;FW_INLINE	= 0x01	; word can be inlined


;;;;;;;;;;;;;;;;;;;;;;;;;;;; Absolute jump macros ;;;;;;;;;;;;;;;;;;;;;;;;;;;;
define({jmpa},{
ifelse({$2},,
	{push #$1
	 ret},
	{mov $2,#$1
	 jmp $2}
)})
define({calla},{
ifelse({$2},,
	{mov eax,#$1
	 call eax},
	{mov $2,#$1
	 call $2}
)})


;;;;;;;;;;;;;;;;;;;;;;; Aerial (extra-light) threads ;;;;;;;;;;;;;;;;;;;;;;;;;
Struc({AT},{
; AerialThread	; These threads are lighter than light: Aerial !
  _long_(SAVE_EBX)	; initial Top of Stack
  _long_(SAVE_ESI)	; Data Stack Pointer
  _long_(SAVE_EDI)	; Return Stack Pointer
  _long_(SAVE_EBP)	; Those two are reserved for the
  _long_(SAVE_ESP)	;; threading mechanism.
  _long_(SAVE_RESP)	; ESP save when going ret-threaded
  _long_(STATE)		; flags used for thread sync'
  _long_(MAX_ESI)	; bottom of data stack
  _long_(MAX_EDI)	; bottom of return stack
  _long_(LINKBACK)	; Link Back to thread object
})
